local log = Winning_Log

-------------------------------------------------------------------------------
-- WinningBgData
--
-- Encapsulates all Blizzard calls regarding BG data.
-------------------------------------------------------------------------------

WinningBgData = {}
WinningBgData.__index = WinningBgData

WinningBgData.new = function(self, player_name)
    local self = {}
    setmetatable(self, WinningBgData)

    self._player_name = player_name
    local found_player = false
    for idx = 1, GetNumBattlefieldScores() do
        local name, killingBlows, honorableKills, deaths, honorGained, faction, race, class, classToken, damageDone, healingDone, bgRating, ratingChange, preMatchMMR, mmrChange, talentSpec = GetBattlefieldScore(idx)
        if name == self._player_name then
            found_player = true
            self._kbs = killingBlows
            self._hks = honorableKills
            self._deaths = deaths
            self._damage = damageDone
            self._healing = healingDone
            self._class = classToken
            self._talent_spec = GetSpecializationInfo(GetSpecialization())
            self._race = race
            self._bg_faction = faction
            self._winning_faction = GetBattlefieldWinner() or 1
            self._bg_date = time()
            self._duration = math.floor(GetBattlefieldInstanceRunTime()/1000)

            SetMapToCurrentZone()
            self._map_id = GetCurrentMapAreaID()

            self._season = GetCurrentArenaSeason()
            self._is_rated = IsRatedBattleground()
        end
    end
    if not found_player then
        log("Player name is: " .. tostring(self._player_name))
        error("Could not find player in BG data: " .. self._player_name)
    end

    return self
end

WinningBgData.get_player_name = function(self)
    return self._player_name
end

WinningBgData.get_player_spec = function(self)
    return self._talent_spec
end

WinningBgData.get_player_race = function(self)
    return self._race
end

WinningBgData.get_date = function(self)
    return self._bg_date
end

WinningBgData.get_map_id = function(self)
    return self._map_id

end

WinningBgData.get_duration = function(self)
    return self._duration
end

WinningBgData.get_won = function(self)
    return self._bg_faction == self._winning_faction
end

WinningBgData.get_kbs = function(self)
    return self._kbs
end

WinningBgData.get_hks = function(self)
    return self._hks
end

WinningBgData.get_deaths = function(self)
    return self._deaths
end

WinningBgData.get_player_damage = function(self)
    return self._damage
end

WinningBgData.get_player_healing = function(self)
    return self._healing
end

WinningBgData.get_season = function(self)
    return self._season
end

WinningBgData.get_was_rated_match = function(self)
    return self._is_rated
end




-------------------------------------------------------------------------------
-- WinningWiper
--
-- Wipes data according to come context (ALL data, SEASON data, etc.)
-------------------------------------------------------------------------------

WinningWiper = {}

WinningWiper.init = function(self, addon)
    self.addon = addon
    self.db = self.addon.db.char.all_records
    self.bg_db = self.addon.db.char.bg_records
    self.bg_db_data = self.addon.db.char.bg_db_data

    -- self.stats = self.addon.db.char.stats
    self.ALL = "ALL"
    self.MOCK = "MOCK"
    self.SEASON = "SEASON"
    self.PREVIOUS_SEASONS = "PREVIOUS_SEASONS"
    self._allowed_wipe_contexts = {
        [self.ALL] = self._wipe_all,
        [self.MOCK] = self._wipe_mock,
        [self.SEASON] = self._wipe_season,
        [self.PREVIOUS_SEASONS] = self._wipe_previous_seasons,
    }
    return self
end


--
-- Wipe bg record data according to a data context and a value.
--
WinningWiper.wipe = function(self, ctx, val)
    local wipe_ctx = self:_check_ctx(ctx)
    if wipe_ctx then
        -- This functions are not bound, have to call with self.
        return wipe_ctx(self, val)
    else
        -- what else to do here?
        return {}
    end
end


-- check wipe context
WinningWiper._check_ctx = function(self, ctx)
    local wipe_ctx = self._allowed_wipe_contexts[ctx]
    if wipe_ctx then
        log("Using wipe context: " .. ctx)
        return wipe_ctx
    else
        log(format("Unknown data context: %s", ctx))
    end
end


-- Wipe all data.
-- This is the only wipe method that actually wipes all BG data used in the UI,
-- for now - at least until I figure out the best way to handle the data.
WinningWiper._wipe_all = function(self, val)
    self:wipe_db()
    self:wipe_bg_db()
    self:wipe_bg_db_data()
end

WinningWiper.wipe_db = function(self)
    local counter = 0
    for i, v in pairs(self.db) do
        self.db[i] = nil
        counter = counter + 1
    end
    log(format("Wiped %d records.", counter))
end

WinningWiper.wipe_bg_db = function(self)
    local counter = 0
    for i, v in pairs(self.bg_db) do
        self.bg_db[i] = nil
        counter = counter + 1
    end
    log(format("Wiped %d bg records.", counter))
end

WinningWiper.wipe_bg_db_data = function(self)
    local counter = 0
    for i, v in pairs(self.bg_db_data) do
        self.bg_db_data[i] = nil
        counter = counter + 1
    end
    log(format("Wiped %d bg records.", counter))
end

WinningWiper._wipe_mock = function(self)
    local counter = 0
    for i, v in pairs(self.db) do
        if v.PlayerName ~= "Morpork" then
            self.db[i] = nil
            counter = counter + 1
        end
    end
    log(format("Wiped %d records.", counter))

end

--
-- WinningBattlegroundRecord model.
--
-- Just a helper to document what an expected bg model looks like.
WinningBattlegroundRecord = {
    uuid = false,
    PlayerName = false,
    PlayerSpec = false,
    PlayerRace = false,
    Type = false,
    Date = false,
    MapId = false,
    Map = false,
    DurationRaw = false,
    Duration = false,
    Won = false,
    Winner = false,
    KBs = false,
    HKs = false,
    Deaths = false,
    Damage = false,
    DamageRaw = false,
    Healing = false,
    HealingRaw = false,
    Season = false,
    Rated = false
}



-------------------------------------------------------------------------------
-- WinningDataAccessLayer
--
-- Control access to Record data.
-------------------------------------------------------------------------------

WinningDataAccessLayer = {}


WinningDataAccessLayer.init = function(self, addon)
    self.addon = addon

    -- This is the master record of all data.
    self.db = addon.db.char.all_records

    -- This is a munged storage of data, fit for UI components.
    self.bg_db = addon.db.char.bg_records

    -- this is a table containing bg records that are currently being shown in
    -- the UI, so that we can calculate stats from them
    self.bg_db_data = addon.db.char.bg_db_data


    -- Time contexts.
    self.ALL = "All"
    self.SEASON = "This Season"
    self.SIX_HOURS = "Last 6 Hours"

    -- Map contexts.
    self.MAP = "MAP"

    return self
end


WinningDataAccessLayer.get_record = function(self, record_idx)
    return self.bg_db[record_idx]
end

-- save bg data
WinningDataAccessLayer.save = function(self, bg_record)
    for k, v in pairs(WinningBattlegroundRecord) do
        log("checking: " .. k)
        assert(bg_record[k] ~= nil)
    end

    log("Saving BG record into DB.")
    -- I'm probably going to regret this.
    self.db[bg_record.uuid] = bg_record
    --table.insert(self.db, bg_record)

    log("Stuffing BG record into BG DB.")
    self:save_bg_data(bg_record)
end


WinningDataAccessLayer.convert_bg_record = function(self, bg_record)
    local ui_data = {}
    for i, v in pairs(self.addon.records_cols) do
        local val = bg_record[v['name']]
        ui_data[i] = val
    end
    return ui_data
end


-- Resync the data that is curated for the UI
WinningDataAccessLayer.resync = function(self, wiper)
    -- wipe the data stores for the UI
    wiper:wipe_bg_db()
    wiper:wipe_bg_db_data()

    for i, bg_record in pairs(self.db) do
        log("Resyncing record: " .. tostring(i))
        self:save_bg_data(bg_record)
    end
end


WinningDataAccessLayer.save_bg_data = function(self, bg_record)
    if bg_record.uuid == nil then
        bg_record.uuid = Winning_UUID()
    end
    self.bg_db_data[bg_record.uuid] = bg_record
    table.insert(self.bg_db, self:convert_bg_record(bg_record))
end


WinningDataAccessLayer.context_updated = function(self, ctx)
    log("Map context value: " .. tostring(ctx.map.value))
    local map_name, canEnter, isHoliday, isRandom, BattleGroundID, mapDescription = GetBattlegroundInfo(ctx.map.value)
    log("Map context real value: " .. tostring(map_name))

    log("Time context value: " .. tostring(ctx.time.value))

    -- this will be slow :(
    for i, v in pairs(self.bg_db) do
        self.bg_db[i] = nil
    end

    for i, bg_record in pairs(self.db) do
        local map_match = false

        log("Checking Map: " .. tostring(bg_record.Map))
        -- if ctx.map.value == 1 or ctx.map.value == bg_record.MapId then
        if ctx.map.value == 1 or map_name == bg_record.Map then
            map_match = true
        else
            log(format("Map did not match: %s -> %s", map_name, bg_record.Map))
        end

        local time_match = false
        if ctx.time.context == self.ALL then
            time_match = true
        elseif ctx.time.context == self.SEASON then
            if bg_record.Season == ctx.time.value then
                time_match = true
            else
                log(format("TIME: Season did not match: %s", bg_record.Season))
            end
        elseif ctx.time.context == self.SIX_HOURS then
            log("Checking Six Hours time context.")
            log(format("%s >= %s", bg_record.Date, ctx.time.value))
            if bg_record.Date >= ctx.time.value then
                time_match = true
            else
                log(format("TIME: Six Hours did not match: %s", bg_record.Date))
            end
        else
            log("TIME: No context match for type? " .. ctx.time.context)
        end

        if map_match and time_match then
            self.bg_db_data[bg_record.uuid] = bg_record
            table.insert(self.bg_db, self:convert_bg_record(bg_record))
        else
            self.bg_db_data[bg_record.uuid] = nil
        end
    end

    for i, v in pairs(self.bg_db) do
        log(format("BG DB Contains: %s -> %s", i, v[2]))
    end

    self.addon:records_updated()
end



-- Max of a particular statistic for a given context.
WinningDataAccessLayer.max_stat_in_ctx = function(self, stat_key)
    local val = 0
    for uuid, bg_record in pairs(self.bg_db_data) do
        local new_val = bg_record[stat_key] or 0
        if new_val > val then
            val = new_val
        end
    end
    return val
end

-- Sum up all of a particular statistic for a given context.
WinningDataAccessLayer.summed_stat_in_ctx = function(self, stat_key)
    local val = 0
    for uuid, bg_record in pairs(self.bg_db_data) do
        local new_val = bg_record[stat_key] or 0
        val = val + new_val
    end
    return val
end


WinningDataAccessLayer.win_loss = function(self)
    local wins = 0
    local losses = 0
    for uuid, bg_record in pairs(self.bg_db_data) do
        if bg_record['Won'] then
            wins = wins + 1
        else
            losses = losses + 1
        end
    end
    return wins, losses
end


WinningDataAccessLayer.win_loss_ratio = function(self, wins, losses)
    if wins == 0 then
        return "0%"
    end
    if losses == 0 and wins > 0 then
        return "100%"
    end
    if wins > 0 and losses > 0 then
        return format("%d%%", math.floor((wins/(wins+losses))*100))
    end
end
