--[[

Winning

Show win/loss records for the following:
- by map
- by day/session
- by week
- by season
- since wipe
- by expac?

TODO: Make records be account/faction bound instead of just character.

For each BG, keep track of:

- player character name
- player character spec
- player character race
- date
- map
- match duration
- match outcome (win/loss)
- honor kills
- killing blows
- deaths
- damage
- healing
- final score? -- will require more work with GetBattle
- rage quits?
- faction played on?

]]


local addon_name = "Winning"
local chat = Winning_Chat
local log = Winning_Log

local bg_record = WinningBattlegroundRecord

Winning = LibStub("AceAddon-3.0"):NewAddon(addon_name, "AceEvent-3.0", "AceConsole-3.0")
local addon = Winning
addon.version = "0.1.1"
addon.context_items = {
    "All",
    "This Season",
    "Last 6 Hours",
}
addon.map_items = {}


addon.ScrollingTable = LibStub("ScrollingTable");


local L = LibStub("AceLocale-3.0"):GetLocale(addon_name)
local addonLDB = LibStub("LibDataBroker-1.1"):NewDataObject(addon_name, {
	type = "data source",
	text = addon_name,
    icon = "Interface\\PvPRankBadges\\PvPRank12",
	OnClick = function(self, btn) Winning_Minimap_Click(btn) end,
});
local icon = LibStub("LibDBIcon-1.0")
local ui_tabs = {'RECORDS', 'STATS'}
local grey_cell_bg = { ["r"] = 0.15, ["g"] = 0.15, ["b"] = 0.15, ["a"] = 1.0 }
local blue = "\124cFF00A9FF"
local red = "\124cFFFF141D"



-------------------------------------------------------------------------------
-- Data Structures For Scrolling Tables
-------------------------------------------------------------------------------

addon.records_cols = {
    {
        ["name"] = L["Date"],
        ["width"] = 150,
        ["align"] = "CENTER",
        ["defaultsort"] = "dsc",
        ["comparesort"] = function (self, rowa, rowb, sortby)
            local rowa = self:GetRow(rowa);
            local rowb = self:GetRow(rowb);

            -- God damn this lib sucks.
            local column = self.cols[sortby]
            local sort_dir = column.sort or column.defaultsort or "asc"

            if sort_dir == 'asc' then
                return rowa[1] > rowb[1]
            else
                return rowa[1] < rowb[1]
            end

        end,
        ["DoCellUpdate"] = function(rowFrame, cellFrame, data, cols, row, realrow, column, fShow, table, ...)
            -- ScrollingTable sucks.
            local record = table:GetRow(realrow)
            cellFrame.text:SetText(date("%a %H:%M %x", record[1]));
        end
    },{
        ["name"] = L["Map"],
        ["width"] = 150,
        ["align"] = "CENTER",
        ["defaultsort"] = "dsc",
    },{
        ["name"] = L["Duration"],
        ["width"] = 60,
        ["align"] = "CENTER",
        ["defaultsort"] = "dsc",
    },{
        ["name"] = L["Winner"],
        ["width"] = 60,
        ["align"] = "CENTER",
        ["defaultsort"] = "dsc",
    },{
        ["name"] = L["KBs"],
        ["width"] = 40,
        ["align"] = "CENTER",
        ["defaultsort"] = "dsc",
    },{
        ["name"] = L["HKs"],
        ["width"] = 40,
        ["align"] = "CENTER",
        ["defaultsort"] = "dsc",
    },{
        ["name"] = L["Deaths"],
        ["width"] = 50,
        ["align"] = "CENTER",
        ["defaultsort"] = "dsc",
    },{
        ["name"] = L["Damage"],
        ["width"] = 60,
        ["align"] = "RIGHT",
        ["defaultsort"] = "dsc",
    },{
        ["name"] = L["Healing"],
        ["width"] = 60,
        ["align"] = "RIGHT",
        ["defaultsort"] = "dsc",
    }
}

for i, col in pairs(addon.records_cols) do
    if i % 2 == 0 then
        col['bgcolor'] = grey_cell_bg
    end
end


-------------------------------------------------------------------------------
-- ACE Framework special methods.
-------------------------------------------------------------------------------

function addon:OnInitialize()
    log("Addon initializing.")

    -- Initialize database
    self.db = LibStub("AceDB-3.0"):New(addon_name .. "Db", {
		profile = {
			minimap = {
				hide = false,
			},
		},
        char = {
            -- Table used for holding data for the UI
            bg_records = {},

            -- Master record of all BG data.
            all_records = {},

            -- Table used for calculating currently visible stats.
            bg_db_data = {},

            time_context_value = 1,
            map_context_value = 1
        }
	});



    addon.dal = WinningDataAccessLayer:init(addon)
    addon.wiper = WinningWiper:init(addon)

    addon:setup_player_faction()

    addon:setup_ui()

	icon:Register(addon_name, addonLDB, self.db.profile.minimap)
	self:RegisterChatCommand("winning", "SlashCommand")
end


function addon:OnEnable()
    self:RegisterEvent("ADDON_LOADED")
    self:RegisterEvent("PLAYER_ENTERING_WORLD")
    self:listen_for_bg_events()
    self:setupDisplay()
end


function addon:OnDisable()
    Winning_UI:Hide()
end


-- Events

function addon:ADDON_LOADED()
    log("Addon loaded.")
    self:setupDisplay()
end


function addon:PLAYER_ENTERING_WORLD()
    self:listen_for_bg_events()
    for i=1,GetNumBattlegroundTypes() do
        if i ~= 1 then
            local bg_name = GetBattlegroundInfo(i)
            addon.map_items[i] = bg_name
        else
            addon.map_items[1] = "All"
        end
    end

end


function addon:listen_for_bg_events()
    self:RegisterEvent("UPDATE_BATTLEFIELD_SCORE")
    self:RegisterEvent("UPDATE_BATTLEFIELD_STATUS")
    self:release_process_token()
end


function addon:ignore_bg_events()
    self:UnregisterEvent("UPDATE_BATTLEFIELD_SCORE")
    self:UnregisterEvent("UPDATE_BATTLEFIELD_STATUS")
end


function addon:UPDATE_BATTLEFIELD_STATUS()
    log("UPDATE_BATTLEFIELD_STATUS")
    self:record_battleground_stats_if_over()
end


function addon:UPDATE_BATTLEFIELD_SCORE()
    log("UPDATE_BATTLEFIELD_SCORE")
    self:record_battleground_stats_if_over()
end



-------------------------------------------------------------------------------
-- Addon functionality
-------------------------------------------------------------------------------

function addon:SlashCommand(args)
    log("Slash command called: " .. tostring(args))

    if args == 'minimap' then
        self.db.profile.minimap.hide = not self.db.profile.minimap.hide
        if self.db.profile.minimap.hide then
            icon:Hide(addon_name)
        else
            icon:Show(addon_name)
        end
    elseif args == 'debug' then
        WINNING_DEBUG = not WINNING_DEBUG
    elseif args == 'stubs' then
        for i = 1, 50 do
            self.dal:save(self:mock_bg_record(i))
        end
    elseif args == 'stubs5' then
        for i = 1, 5 do
            self.dal:save(self:mock_bg_record(i))
        end
    elseif args == 'stubs500' then
        for i = 1, 500 do
            self.dal:save(self:mock_bg_record(i))
        end
    elseif args == 'stubs5000' then
        for i = 1, 5000 do
            self.dal:save(self:mock_bg_record(i))
        end
    elseif args == 'wipe-mock' then
        self.wiper:wipe(addon.wiper.MOCK, true)
    elseif args == 'wipe-all' then
        self.wiper:wipe(addon.wiper.ALL, true)
    elseif args == 'wipe-season' then
        self.wiper:wipe(addon.wiper.SEASON, current_season)
    elseif args == 'wipe-previous' then
        self.wiper:wipe(addon.wiper.PREVIOUS_SEASONS, current_season)
    elseif args == 'resync' then
        self.dal:resync(self.wiper)
    elseif args == 'show' then
        self:toggle_ui()
    else
        chat("Available commands: minimap, debug, wipe-all, show" )
    end

    self:refresh()
end



function addon:request_process_token()
    if self._bg_process_token == nil then
        self:ignore_bg_events()
        self._bg_process_token = time()
    end
    return self._bg_process_token
end



function addon:release_process_token()
    self._bg_process_token = nil
end



--
-- Record If Over
--
function addon:record_battleground_stats_if_over()
	local _, zone_type = IsInInstance()
    if zone_type == "pvp" then
        local bg_winning_faction = GetBattlefieldWinner()
        if bg_winning_faction ~= nil then
            if self:request_process_token() ~= nil then
                -- If saving a BG record breaks, the addon should still function
                local player_name = UnitName("player")
                local success, result = pcall(self.process_battleground_stats, self, player_name)
                if success then
                    log("Successfully recorded battleground stats.")
                else
                    log("Error recording battleground stats.")
                    log(tostring(result))
                end
            else
                log("Match over, but could not get token. Ignoring.")
            end
        else
            log("Match not over yet.")
        end
    end
end


--
-- Process BG Stats for current player.
--
function addon:process_battleground_stats(player_name)
    bg = WinningBgData:new(player_name)
    local bg_record = {}
    bg_record.uuid = Winning_UUID()
    bg_record.PlayerName = bg:get_player_name()
    bg_record.PlayerSpec = bg:get_player_spec()
    bg_record.PlayerRace = bg:get_player_race()
    bg_record.Type = 'bg_record'
    bg_record.Date = bg:get_date()
    bg_record.MapId = bg:get_map_id()
    bg_record.Map = self:render_bg_map(bg_record.MapId)
    bg_record.DurationRaw = bg:get_duration()
    bg_record.Duration = self:render_bg_duration(bg_record.DurationRaw)
    bg_record.Won = bg:get_won()
    bg_record.Winner = self:render_bg_won(bg_record.Won)
    bg_record.KBs = bg:get_kbs()
    bg_record.HKs = bg:get_hks()
    bg_record.Deaths = bg:get_deaths()
    bg_record.DamageRaw = bg:get_player_damage()
    bg_record.Damage = self:render_damage(bg_record.DamageRaw)
    bg_record.HealingRaw = bg:get_player_healing()
    bg_record.Healing = self:render_healing(bg_record.HealingRaw)
    bg_record.Season = bg:get_season()
    bg_record.Rated = bg:get_was_rated_match()

    self.dal:save(bg_record)
end


function addon:mock_bg_record(i)
    local map = math.random(2, 13)
    local map_name = GetBattlegroundInfo(map)
    local winner = math.random(0, 1)
    local won = winner == self.player_faction_num
    local duration = math.random(300, 1200)
    local damage = math.random(100000, 9999999)
    local healing = math.random(100000, 9999999)

    local _, race = UnitRace("player")

    return {
        uuid = Winning_UUID(),
        PlayerName = 'foo' .. tostring(i),
        PlayerSpec = 102, -- balance druid
        PlayerRace = race,
        Type = 'bg_record',
        Date = time() + math.random(1000, (1000*i)+1) - 1000000000,
        MapId = map,
        Map = map_name,
        DurationRaw = duration,
        Duration = self:render_bg_duration(duration),
        Won = won,
        Winner = addon:render_bg_won(won),
        KBs = math.random(1, 30),
        HKs = math.random(1, 100),
        Deaths = math.random(1, 10),
        Damage = self:render_damage(damage),
        DamageRaw = damage,
        Healing = self:render_healing(healing),
        HealingRaw = healing,
        Season = 1,
        Rated = false,
    }
end

function addon:setup_player_faction()
    self.player_faction = UnitFactionGroup("player");
    if self.player_faction == "Neutral" then
        self.player_faction_num = nil
    else
        self.player_faction_num = 1
        if self.player_faction == "Horde" then
            self.player_faction_num = 0
        end
    end
    log("Player faction is: " .. self.player_faction)
    log("Player faction num is: " .. tostring(self.player_faction_num))
end





function addon:render_bg_won(won)
    -- Horde team always Red
    -- Alliance team always Blue
    -- text for opposing team depends on what faction the player was playing as

    local alliance = self:blue(FACTION_ALLIANCE)
    local horde = self:red(FACTION_HORDE)

    if won then
        if self.player_faction_num == 0 then
            return horde
        else
            return alliance
        end
    else
        -- if the player did not win, always return opposite faction in that
        -- faction's color.
        if self.player_faction_num == 0 then
            return alliance
        else
            return horde
        end
    end
end


function addon:render_bg_duration(total_seconds)
    return Winning_Duration(total_seconds)
end


function addon:blue(txt)
    return format("%s%s\124r", blue, txt)
end


function addon:red(txt)
    return format("%s%s\124r", red, txt)
end


function addon:render_bg_map(map)
    return GetMapNameByID(map)
end


function addon:render_damage(damage)
    return self:_humanify(damage)
end

function addon:render_healing(healing)
    return self:_humanify(healing)
end


function addon:_humanify(num)
    local adjusted = 0
    if num > 999999999 then
        -- round billions
        return format("%.1fB", num/1000000000)
    elseif num > 999999 then
        -- round millions
        return format("%.2fM", num/1000000)
    elseif num > 999 then
        -- round thousands
        return format("%dk", math.ceil(num/1000))
    end
    return num
end



-------------------------------------------------------------------------------
-- UI Functions
-------------------------------------------------------------------------------

-- one-time setup of the UI
function addon:setup_ui()
    log("Setting up info.")

    Winning_UI_Info_Ratio_Label:SetText(L['WINNING'])

    Winning_UI_Info_Stats_HKs_Label:SetText(L['Top HKs'])
    Winning_UI_Info_Stats_KBs_Label:SetText(L['Top KBs'])
    Winning_UI_Info_Stats_Damage_Label:SetText(L['Damage'])
    Winning_UI_Info_Stats_Healing_Label:SetText(L['Healing'])

    if self.records_table == nil then
        self.records_table = addon.ScrollingTable:CreateST(addon.records_cols, 25, nil, nil, Winning_UI_Records_Table)
        log("BG DB: " .. tostring(self.dal.bg_db))
        self.records_table:SetData(self.dal.bg_db, true)
        log("ST (data): " .. tostring(self.records_table.data))
    end
    -- Now that the table is instantiated, fit it to it's parent frame.
    _G[self.records_table["frame"]:GetName()]:SetAllPoints();

    -- setup context dropdown
    UIDropDownMenu_Initialize(Winning_Context_DropDown, Winning_Context_DropDown_OnLoad);
    UIDropDownMenu_SetWidth(Winning_Context_DropDown, 100);
    UIDropDownMenu_SetButtonWidth(Winning_Context_DropDown, 100);
    UIDropDownMenu_SetSelectedID(Winning_Context_DropDown, 1);
    UIDropDownMenu_JustifyText(Winning_Context_DropDown, "LEFT");

    UIDropDownMenu_Initialize(Winning_Map_DropDown, Winning_Map_DropDown_OnLoad);
    UIDropDownMenu_SetWidth(Winning_Map_DropDown, 100);
    UIDropDownMenu_SetButtonWidth(Winning_Map_DropDown, 100);
    UIDropDownMenu_SetSelectedID(Winning_Map_DropDown, 1);
    UIDropDownMenu_JustifyText(Winning_Map_DropDown, "LEFT");
end


-- Used to refresh the Winning UI.
function addon:refresh()
    self:updateRecordsTab()
    self:updateInfo()
end


function addon:updateInfo()
    log("Updating info box.")

    local wins, losses = self.dal:win_loss()
    local win_loss_ratio = self.dal:win_loss_ratio(wins, losses)

    Winning_UI_Info_Ratio_Wins:SetText(wins)
    Winning_UI_Info_Ratio_Losses:SetText(losses)
    Winning_UI_Info_Ratio_Value:SetText(win_loss_ratio)

    Winning_UI_Info_Stats_HKs_Value:SetText(
        self.dal:max_stat_in_ctx("HKs")
    )
    Winning_UI_Info_Stats_KBs_Value:SetText(
        self.dal:max_stat_in_ctx("KBs")
    )
    Winning_UI_Info_Stats_Damage_Value:SetText(self:render_damage(
        self.dal:max_stat_in_ctx("DamageRaw")
    ))
    Winning_UI_Info_Stats_Healing_Value:SetText(self:render_healing(
        self.dal:max_stat_in_ctx("HealingRaw")
    ))
end



function addon:updateRecordsTab()

    -- TODO: How to handle dynamic filters?
    local filterFunction = function(self, row)
        return true
    end
    self.records_table:SetFilter(filterFunction);
end



-- Called after the DAL is done updating the data according to the context.
function addon:records_updated()
    log("Records updated, sorting table.")
    self:updateInfo()
    self.records_table:SortData()
end


-- called when the context dropdowns are updated.
function addon:context_updated()
    self.dal:context_updated(self:current_context())
end


function addon:current_context()
    local ctx = {}

    -- Time frame context
    local time_selected_id = Winning_Context_DropDown['selectedID']
    local time_ctx = addon.context_items[time_selected_id]

    local time_value = self.dal.ALL
    if time_ctx == self.dal.SEASON then
        time_value = GetCurrentArenaSeason()
    elseif time_ctx == self.dal.SIX_HOURS then
        time_value = time() - (60*60*6)
    end

    log("Time ctx of: " .. tostring(time_ctx))
    log("Time value of: " .. tostring(time_value))

    ctx.time = {}
    ctx.time.context = time_ctx
    ctx.time.value = time_value


    -- Map context
    local map_selected_id = Winning_Map_DropDown['selectedID']
    local map_ctx = addon.map_items[map_selected_id]

    log("Map ctx of: " .. tostring(map_ctx))
    ctx.map = {}
    ctx.map.value = map_selected_id

    return ctx
end


function addon:setupDisplay()

    log("Setting up display.")

    -- Initialize all text and things
    Winning_UI_Title:SetText(format("%s v%s", addon_name, addon.version))

    Winning_UITab1:SetText(L['Records'])
    Winning_UITab2:SetText(L['Stats'])

    Winning_UI_Info_Context_Time_Label:SetText(L['Context'])
    Winning_UI_Info_Context_Maps_Label:SetText(L['Maps'])

    Winning_Info_Report_Label:SetText(L['Report'])
    Winning_UI_Report_Party:SetText(L['Party'])
    Winning_UI_Report_Guild:SetText(L['Guild'])
end


function addon:resetPosition()
    log("Resetting UI position.")
    ui:ClearAllPoints()
    ui:SetPoint("CENTER", UIParent, "CENTER")
end


function addon:updateDisplay()
    log("Updating UI display.")
end


function addon:toggle_ui()
    log("toggling ui.")
    if Winning_UI:IsVisible() then
        -- hide main UI
        log("Hiding main UI.")
        Winning_UI:Hide()
    else
        -- show main UI
        log("Showing main UI.")
        Winning_UI:Show();
        addon:updateDisplay()
    end
end


-- Used by the reporting buttons at the top of the UI
-- 
-- ALL BGS:
-- 43 Wins
-- 34 Losses
-- WINNING: 55%
-- 
-- LAST SIX HOURS:
-- 43 Wins
-- 34 Losses
-- WINNING: 55%
--
-- CURRENT SEASON:
-- 43 Wins
-- 34 Losses
-- WINNING: 55%
--
function Winning_Report(channel)
    log("Reporting to channel: " .. channel)

    local ctx = addon:current_context()
    local icon = "{skull}"

    local time_ctx = "(unknown)"
    if ctx.time.context == addon.dal.ALL then
        time_ctx = "ALL TIME"
    elseif ctx.time.context == addon.dal.SEASON then
        time_ctx = "THIS SEASON"
    elseif ctx.time.context == addon.dal.SIX_HOURS then
        time_ctx = "LAST SIX HOURS"
    end

    local map_ctx = nil
    if ctx.map.value ~= 1 then
        map_ctx = format("MAP: %s", addon.map_items[ctx.map.value])
    end

    local wins, losses = addon.dal:win_loss()
    local win_loss_ratio = addon.dal:win_loss_ratio(wins, losses)

    if channel ~= "GUILD" then
        SendChatMessage(format("%s %s %s", icon, time_ctx, icon), channel, nil, nil)
        if map_ctx then
            SendChatMessage(map_ctx, channel, nil, nil)
        end
        SendChatMessage(format("%s Wins", wins), channel, nil, nil)
        SendChatMessage(format("%s Losses", losses), channel, nil, nil)
        SendChatMessage(format("Winning %s", win_loss_ratio), channel, nil, nil)
    end
end


-- Used to show contents of a single tab.
function Winning_Show_Tab(tab_name)
    if tab_name == ui_tabs[1] then
        log("Showing tab RECORDS")
        Winning_UI_Stats:Hide()
        Winning_UI_Records:Show()
    elseif tab_name == ui_tabs[2] then
        log("Showing tab STATS")
        Winning_UI_Records:Hide()
        Winning_UI_Stats:Show()
    end
    Winning_UI_Info:Show()
end


-- Called when the tabs at the bottom of the UI are clicked.
function Winning_Tab_Clicked(tab_name)
    log("Tab clicked: " .. tab_name)
    for i, tab in ipairs(ui_tabs) do
        if tab == tab_name then
            PanelTemplates_SetTab(Winning_UI, i)
            Winning_Show_Tab(tab_name)
        end
    end
end


-- Called when the minimap button is clicked.
function Winning_Minimap_Click(btn)
    log(format("Minimap clicked: %s", tostring(btn)))
	if btn == "LeftButton" or btn == nil then
        addon:toggle_ui()
	end
end


-- Handles setting up the context dropdown.
function Winning_Context_DropDown_OnLoad(self, level)
    for i, v in pairs(addon.context_items) do
        local info = UIDropDownMenu_CreateInfo()
        info.text = L[v]
        info.func = Winning_Context_DropDown_OnClick
        UIDropDownMenu_AddButton(info, level);
    end
end


function Winning_Context_DropDown_OnClick(self)
	UIDropDownMenu_SetSelectedID(Winning_Context_DropDown, self:GetID());
    local selectedId = Winning_Context_DropDown['selectedID']
    local ctx = addon.context_items[selectedId]
    addon.db.char.time_context_value = selectedId
    log("time context selected: " .. tostring(ctx))
    addon:context_updated()
end


-- Handles setting up the maps dropdown.
function Winning_Map_DropDown_OnLoad(self, level)
    for i, v in pairs(addon.map_items) do
        local info = UIDropDownMenu_CreateInfo()
        log("creating dropdown info: " .. tostring(v) .. tostring(i))
        info.text = v
        info.value = i
        info.func = Winning_Map_DropDown_OnClick
        UIDropDownMenu_AddButton(info, level);
    end
end

function Winning_Map_DropDown_OnClick(self)
	UIDropDownMenu_SetSelectedID(Winning_Map_DropDown, self:GetID());
    local selectedId = Winning_Map_DropDown['selectedID']
    local ctx = addon.map_items[selectedId]
    addon.db.char.map_context_value = selectedId
    log("Map context selected: " .. tostring(ctx))
    addon:context_updated()
end
